!+
! Module: rgssources_mod
!
! Authors:	Peter Videler - ESTEC/SOC
!		Ian Stewart - SSC/LUX
!		Masaaki Sakano - SSC/LUX
!          
! Description: Construct table with source positions to be used in RGS pipeline.
! 
! Comments:
!
!	TODO:
!	- INSTRUME -> M1 for RGS1, M2 for RGS2, if possible.
!	- apply a loose spatial filter for the first EPIC source selection before calculating confusion.
!	- write the section of Attitude parameters in the reference.
!	- test old list for empty labels?
!	- even more rigorous testing of old list??
!	- quit under any circumstances if old rgs list doesn't pass checks.
!	- write EXPOSURE kwd - see JO mail of 25 April.
!	- can we get rid of the necessity to have --changeprime=yes as well as --withusersource=yes and userasprime=yes?
!	- take out dataMode=spectroscopy from oal set state (or elsewhere) and see what happens. (See CE mail of 27 July.)
!	- what about updates of post-5.1 files which nevertheless have not the full set of header keywords, or wrong disp/xdsp, etc?
!	- Change column names in kwds (eg FLAG* written to the table by dataFilter) in cases where the old name no longer exists?
!	- in checks of old file, make sure prime src index is sensible. User might have deleted rows or altered the number.
!	- Allow the task to extract the instrument and expsoure from rgsproc-format file names.
!	- When primestyle=label, it can't find the default PROPOSAL. String dump to check for trailing nulls etc.
!	- EXP_ID keyword will have the wrong value (3-digit exp str instead of 13-digit obs+exp str) for a lot of pre-sas_5.3 rgs source lists. Change the update module so it patches these instances up.
!	- Rationalise --epicset (and srclist??) parameters - should they both be of type t instead of e?
!	- ras, decs and pas - parameters should be of type u (angle) but need to wait until the param interface is updated so that default settings can be properly identified.
!	- See mail JDR 12 Mar 2002. Split xdsp 'slit' width selection into 2: one of +-3 arcmin (or, even better, specified by a user parameter) to select sources; the other smaller, intended for pcms use, to make sure that prime source always lies right on the chip.
!
! $Id: rgssources_mod.f90,v 1.5 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use errorHandling
  use param  
  use oal		! includes 'use caloaldefs'
  use dataFilter	! in selectlib
  use odfdesignators	! In caloalutils. Needed to dissect instexpid.


  use pointing_mod	! -> SkyPositionType
  use dummycal_mod

  use array_utils
  use constants
!!use temp_ssclib
  use confusion		! in ssc_misc
  use coordinate
  use psf_ims
  use ssc_misc		! for intToStr, ucase.

  use rgssources_init	! -> sourceType
use rgssources_shared	! If rgssources_update ever becomes a separate task,
use rgssources_update	! all the stuff in shared should go back -> main.

  implicit none 

  character(255), parameter :: epicBandsConfFileName = "epicenergybands.conf"
  character(255) :: fovFilterStr
  type(SpacecraftAttitudeType) :: scAttitude, rgsAttitude
  ! type defined in caloalutils/src/caloaldefs.f90
  integer(int32) ::&
    paramInstrumentId
  character(4) ::&
    paramExposureIdStr
!  real(single), parameter ::&
!    detUnitArcsec = 0.05 ! The unit of the DETX,Y system.
!!*** should be in constants module
  real(double) ::&
    xPixelSizeWorld,&
    yPixelSizeWorld,&
    pixelAreaArcsec
  logical :: haveScAttitude

! Command-line parameters:
  logical ::&
    withEpicSet = .false.,&
    addUserSrc = .false.,&
    changePrimeSrc = .false.,&
    changeAttitude = .false.,&
    clobberOnLabelClash = .false.,&
    setUserSrcToPrime = .false.,&
    enableSpatialFilter = .false.,&
    writeObsAttributes = .false.,&
    writeExpAttributes = .false.,&
    processUserSrc = .false.,&
    bkgExcludeUserSrc = .false.,&
    doEpicConfusion = .false.,&
    doFlagEpicSrcOutOfFov = .false.
  character(len=256) ::&
    oldListSetName = '',&
    newListSetName = '',&
    epicSrcSetName = '',&
    atthkSetName = '',&
    primeSrcExpression = '',&
    paramInstExpIdStr = '',&
    attfilterTableName = ''
!,&
!    tempSrcSet = ''
  character(len=9) ::&
    userSrcStyle = '',&
    primeSrcStyle = '',&
    attitudeStyle = ''
  character(len=20) ::&
    fileMode = ''
  character(len=labelsize) ::&
    userSrcLabel = '',&
    primeSrcLabel = ''
  character(len=labelsize-5) ::&
    epicLabelPrefix = ''
  real(single) ::&
    userSrcRate = 0.0,&
    userSrcRaDeg = 0.0,&
    userSrcDecDeg = 0.0,&
    userSrcXdspArcmin = 0.0,&
    userSrcDispArcmin = 0.0,&
    userAttitudeRaDeg = 0.0,&
    userAttitudeDecDeg = 0.0,&
    userAttitudeAposDeg = 0.0,&
    bandMinEeV = 0.0,&
    bandMaxEeV = 0.0,&
    fracBackground = 0.0
  integer(int16) ::&
    paramIndexOfPrime = 0
  type(instWeightType), allocatable ::&
    weights(:)

!integer :: devId
!real :: polyx(4), polyy(4)

  private 

  public rgssources

!logical :: testFlag = .true.

contains 

  !---------------------------------------------------------------------
  subroutine rgssources
  !---------------------------------------------------------------------

!use caloaldefs
    implicit none 

    integer(int16) :: maxNumSrcs, actualNumSrcs, newPrimeSrcIndex
    type(sourceType), dimension(:), allocatable :: newListSrcData
    type(DataSetT) :: set
    type(TableT) :: tab
    integer(int32) :: instrumentId, epicInstrumentId, newEpicAttrNum, numRows
    character(len=5) :: instrumentName
    type(ColumnT) ::&
      confnCol
    real(single), dimension(:), pointer ::&
      confnColPtr
    character(255) :: instExpIdStr
    real(double) :: epicTimeStamp

    type(InstrumentExposureIdentifierT) :: paramInstExpIdStructure
    ! type defined in caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start:

    call readParameters
if (paramInstExpIdStr .eq. '') then ! NOTE!!! this codelet relies
! on the default value '' of the parameter --instexpid. If this is
! changed, the codelet WILL BECOME INVALID.
paramInstrumentId = XMM
paramExposureIdStr = '' 
else
    paramInstExpIdStructure&
     = instrumentExposureIdentifier(trim(paramInstExpIdStr))
    paramInstrumentId  = instrument(paramInstExpIdStructure)
    paramExposureIdStr = exposureId(paramInstExpIdStructure) 

    ! type defined in caloalutils/src/caloaldefs.f90
    !                                    RGS1        = 8, RGS2 = 9,
    if ((paramInstrumentId .ne. RGS1) .and. (paramInstrumentId .ne. RGS2)) then
      write(errstr, '(a, i2, a)') 'instexpid gives the instrument number ('&
           , paramInstrumentId, '), which is neither RGS1 nor 2.'
      call warning('oddInstrumentId')
    end if
end if

    if (trim(fileMode) .eq. 'modify') then
      call convertOldFiles(&
        oldListSetName,&
        writeObsAttributes,&
        writeExpAttributes,&
        paramInstrumentId,&
        paramInstExpIdStr& 
      )
      call message(AppMsg, Noisy, 'Successfully read the input RGS source list.')
    end if

    call getInstrument(instrumentId, instrumentName)
    call getInstExpId(instrumentId, instExpIdStr)
    call CAL_setState(instrument=instrumentId)
    call getPointing(instrumentId, instExpIdStr) ! sets global variables
    ! scAttitude and rgsAttitude.
    call message(AppMsg, Noisy, 'Successfully read the attitude infomation.')

    call calcMaxNumSrcs(maxNumSrcs)
!    if (enableSpatialFilter) call constructSelectionExpr
if (enableSpatialFilter) fovFilterStr = '(DELTA_XDSP in [-2.9:2.9])'
!fovFilterStr = '(DELTA_XDSP in [-3.0:3.0])'

!.......................................................................
! Construct the source list in memory:

    call message(AppMsg, Noisy, 'Now, constructing the new RGS source list.')
    allocate(newListSrcData(maxNumSrcs))

    call fillOutputSrcList(newListSrcData, actualNumSrcs, newEpicAttrNum)
    ! actualNumSrcs is defined here.

!.......................................................................
! Now write (most of) the source list to file:

    call message(AppMsg, Noisy, 'Now, writing the RGS source list.')
    call writeToDataSet(newListSrcData(1:actualNumSrcs), instrumentName&
    , newEpicAttrNum, instExpIdStr, set, tab)

!.......................................................................
! Last bits and pieces - FOV flagging, prime source, rga-plane confusion,
! and checks:

    ! Flag on whether passes FOV filter or not, then copy the column back
    ! into newListSrcData.
    !
    if (enableSpatialFilter) then
      ! NOTE! Epic sources which fall outside the FOV may have already been
      ! screened out at this stage (if --flagepicsrcoutoffov=false).
      ! Even in that case some sources still have the flag value != 0,
      ! namely proposal, pointing, previous rgs
      ! and user sources. If the pointing is not within the FOV there is
      ! something badly wrong!
      !
      call initFilter('isNull(DELTA_XDSP) || !'//fovFilterStr)	! in dataFilter in selectlib
      numRows = doSelect(tab, 'FLAG', flagBitPos=OUTOFFOV_BIT)	! in dataFilter in selectlib
      call reportOofovFlags(tab, newListSrcData(1:actualNumSrcs)) ! only print warning
!***** warnings on newListSrcData%flag are at present done within this sub - this should really be made more global.
    end if

    if (withEpicSet .and. doEpicConfusion) then
      call getEpicAttData(epicInstrumentId, epicTimeStamp)
      call calcEpicPlaneConfusion(epicInstrumentId, epicTimeStamp&
      , newListSrcData(1:actualNumSrcs))

      ! NOTE!! The cal state is reset to the relevant epic camera inside
      ! this sub.

      call flagConfusedSources(tab, newListSrcData(1:actualNumSrcs))
    end if

    newPrimeSrcIndex = calcPrimeIndex(tab, newListSrcData(1:actualNumSrcs))

    if (trim(fileMode) .eq. 'create' .or. changePrimeSrc) then
      call setAttribute(tab, 'PRIMESRC', newPrimeSrcIndex, COMMPRIMESRC)
    end if

    call calcRgsPlaneConfusion(newListSrcData(1:actualNumSrcs)&
    , newPrimeSrcIndex)

!***** check now that no columns have nulls in them except EPIC_FILE.

    ! Write rgs-plane confusion column to file.
    confnCol = column(tab, 'CONFUSION')
    confnColPtr => real32Data(confnCol)
    confnColPtr = newListSrcData(1:actualNumSrcs)%confusion
    call release(confnCol)

!    call deleteColumn(tab, 'FLAG') ! because there isn't one yet in the disk
!    ! file format.

    call message(AppMsg, Verbose, 'Closing source list.')

    call release(set)

    deallocate(epicBandIds)	! to eliminate the error (see SSC-SPR-3661)
    ! "ALLOCATABLE array EPICBANDIDS has already been ALLOCATEd
    !  Program terminated by fatal error".
    ! This epicBandIds is defined in ./rgssources_init.f90 as a global variable
    ! (int16, allocatable, public) and allocated only once in the
    ! subroutine readParameters below.  Hence, this statement should not be
    ! needed in principle, however for some reason this change fixes the problem
    ! (many thanks to Aitor Ibarra).

  end subroutine rgssources

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: numWeights, numBands, i

    !-------------------------------------------------------------------
    ! Start:

!    tempSrcSet = stringParameter('tempset')

    fileMode = stringParameter('filemode')
    if (trim(fileMode) .eq. 'modify') then
      oldListSetName = stringParameter('srclist')
      changePrimeSrc = booleanParameter('changeprime')
      changeAttitude = booleanParameter('changeattitude')
    else
      newListSetName = stringParameter('srclist')
      changePrimeSrc = .true.
    end if

    paramInstExpIdStr = stringParameter('instexpid')
    writeObsAttributes = booleanParameter('writeobskwds')
    writeExpAttributes = booleanParameter('writeexpkwds')
    clobberOnLabelClash = booleanParameter('clobberonlabel')


    primeSrcStyle = stringParameter('primestyle')
    select case(primeSrcStyle)
    case('auto')
    case('brightest')
    case('index')
      paramIndexOfPrime = intParameter('primeindex')

    case('label')
      primeSrcLabel = ucase(stringParameter('primelabel'))

    case('expression')
      primeSrcExpression = stringParameter('primeexpression')

    case default
call error('badPrimeStyle',"Value of --primestyle not recognized.")
    end select


    attitudeStyle = stringParameter('attstyle')
    select case(attitudeStyle)
    case('mean')
      atthkSetName = stringParameter('meanset')

    case('median')
      atthkSetName = stringParameter('medianset')

    case('start')

    case('user')
      userAttitudeRaDeg   = realParameter('attra')
      userAttitudeDecDeg  = realParameter('attdec')
      userAttitudeAposDeg = realParameter('attapos')

    case('expmedian')
      attfilterTableName = stringParameter('expmediantable')

    case default
call error('badAttStyle',"Value of --attstyle not recognized.")
    end select


    addUserSrc = booleanParameter('addusersource')
    if (addUserSrc) then
      userSrcLabel = ucase(stringParameter('label'))
      userSrcRate = realParameter('rate')
      setUserSrcToPrime = booleanParameter('userasprime')
      processUserSrc    = booleanParameter('process')
      bkgExcludeUserSrc = booleanParameter('bkgexclude')

      userSrcStyle = stringParameter('positionstyle')
      if (userSrcStyle .eq. 'radec') then
        userSrcRaDeg = realParameter('ra')
        userSrcDecDeg = realParameter('dec')
      else ! userSrcStyle .eq. 'wrtatt'
        userSrcDispArcmin = realParameter('deltadisp')
        userSrcXdspArcmin = realParameter('deltaxdsp')
      end if

    else
      setUserSrcToPrime = .false.
    end if

    withEpicSet = booleanParameter('withepicset')
    if (withEpicSet) then
      epicSrcSetName = stringParameter('epicset')
      epicLabelPrefix = stringParameter('epiclabelprefix')
      doEpicConfusion = booleanParameter('doconfusion')
!!!      fracBackground = realParameter('bkgfraction')
fracBackground = 10.0
      numWeights = parameterCount('instweights')
      allocate(weights(numWeights))
      do i = 1, numWeights
        weights(i)%idInst = i
        weights(i)%weight = realParameter('instweights', i-1)
      end do
      doFlagEpicSrcOutOfFov = booleanParameter('flagepicsrcoutoffov')
    end if

    enableSpatialFilter = booleanParameter('enablefilter')

    numBands = parameterCount('bandids')
    allocate(epicBandIds(numBands))
    do i = 1, numBands
      epicBandIds(i) = intParameter('bandids', i-1)
    end do
    if (numBands > 9) then
      call warning('tooManyEpicBandIds'&
           , 'Number of bandids specified is more than 9, which will lead &
           &to missing information of some band-IDs in the header in the &
           &output RGS source list.')
    end if

!    call openParameters(trim(epicBandsConfFileName))
!if (parameterCount('bandloev') .lt. 2 .or. parameterCount('bandhiev') .lt. 2)&
!call error('tooFewEpicBands', 'Too few band limits found in the file '&
!//trim(epicBandsConfFileName)//'.')
!    bandMinEeV = realParameter('bandloev', epicBandNum - 1)
!    bandMaxEeV = realParameter('bandhiev', epicBandNum - 1)
!    call closeParameters()
!*******
bandMinEeV = 500.0		! hard-coded for the use for the PSF
bandMaxEeV = 2000.0		! for calculating calcEpicPlaneConfusion().

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getInstrument(instrumentId, instrumentName)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(out) :: instrumentId
    character(*),   intent(out) :: instrumentName

    type(DataSetT) :: oldListSet

    !-------------------------------------------------------------------
    ! Start:

    if (trim(fileMode) .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      instrumentName = stringAttribute(oldListSet, 'INSTRUME')
      call release(oldListSet)
      instrumentId = instId(instrumentName)

      if (instrumentId .ne. paramInstrumentId) then
        call warning('instIdClash'&
        , "The instrument "//trim(instrumentName)//" recorded in the source &
        &list header keyword &INSTRUME is not the same as that derived from &
        &the value of the --instexpid parameter. Note: if you &
        &have left the --instexpid parameter at its (nonsense) default &
        &setting, then you needn't be alarmed - currently the code has no &
        &way to distinguish between this situation and the one in which &
        &the user _has_ set the parameter but has given it the wrong value.")
      end if

    else
      instrumentId = paramInstrumentId
      instrumentName = instId(instrumentId)
    end if

  end subroutine getInstrument

  !---------------------------------------------------------------------
  subroutine getInstExpId(instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(in) :: instrumentId
    character(*), intent(out) :: instExpIdStr

    type(DataSetT) :: oldListSet
    character(4) :: exposureIdStr

    !-------------------------------------------------------------------
    ! Start:

    if (trim(fileMode) .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      exposureIdStr = stringAttribute(oldListSet, 'EXPIDSTR')
      call release(oldListSet)

      if (exposureIdStr .ne. paramExposureIdStr) then
        call warning('expIdStrClash'&
        , "The 4-character exposure identifier "//exposureIdStr//" recorded &
        &in the source list header keyword EXPIDSTR is not the same as that &
        &derived from the value of the --instexpid parameter. Note: if you &
        &have left the --instexpid parameter at its (nonsense) default &
        &setting, then you needn't be alarmed - currently the code has no &
        &way to distinguish between this situation and the one in which &
        &the user _has_ set the parameter but has given it the wrong value.")
      end if

      if (instrumentId .eq. RGS1) then
        instExpIdStr = 'R1'//exposureIdStr
      else if (instrumentId .eq. RGS2) then
        instExpIdStr = 'R2'//exposureIdStr
      else
call error('bug0', 'Contact the code developer.')
      end if

    else
      instExpIdStr = paramInstExpIdStr
    end if

  end subroutine getInstExpId

  !---------------------------------------------------------------------
  subroutine getPointing(instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

!    use pointing_mod

    implicit none 

    integer(int32), intent(in) ::&
      instrumentId
    character(*), intent(in) :: instExpIdStr

    real(double) ::&
      startTime
    type(DataSetT) :: atthkSet, oldListSet, attfilterSet
    type(TableT) :: attfilterTab
    type(ExposureInfoType) ::&
      exposureInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Obtaining pointing/attitude.')

    haveScAttitude = .false. ! default

    if (trim(fileMode) .eq. 'modify' .and. .not. changeAttitude) then ! get the
    ! attitude from the old rgs source list:
      oldListSet = dataSet(oldListSetName, READ)
      rgsAttitude%ra   = real64Attribute(oldListSet, 'RA_PNT')  * piOn180
      rgsAttitude%dec  = real64Attribute(oldListSet, 'DEC_PNT') * piOn180
      rgsAttitude%apos = real64Attribute(oldListSet, 'PA_PNT')  * piOn180
      call release(oldListSet)

      call reportRgsAttitude()
  return
    end if

    if (attitudeStyle .eq. 'user') then
!    ! NOTE!! The following checks are designed mostly to detect when the user has allowed the attra, attdec and attapos parameters to retain their default values. The user should not do this if attstyle has been set to 'user'. The checks depend to some degree on the values of the relevant parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
!      if (userAttitudeRaDeg .lt. 0.0 .or. userAttitudeRaDeg .gt. 360.0) then
!call error('illegalAttRa', 'The value of --attra, the user-specified &
!&attitude ra, is either < 0  or > 360 degrees. Have you forgotten to set it?')
!      end if
!      if (userAttitudeDecDeg .lt. -90.0 .or. userAttitudeDecDeg .gt. 90.0) then
!call error('illegalAttDec', 'The value of --attdec, the user-specified &
!&attitude declination, is either < -90 or > 90 degrees. Have you forgotten &
!&to set it?')
!      end if
!      if (userAttitudeAposDeg .lt. 0.0 .or. userAttitudeAposDeg .gt. 360.0) then
!call error('illegalAttApos', 'The value of --attapos, the user-specified &
!&attitude position angle, is either < 0  or > 360 degrees. Have you forgotten &
!&to set it?')
!      end if

      rgsAttitude%ra   = userAttitudeRaDeg * piOn180
      rgsAttitude%dec  = userAttitudeDecDeg * piOn180
      rgsAttitude%apos = userAttitudeAposDeg * piOn180

      call reportRgsAttitude()
  return
    end if

!    call OAL_setState(instrument=instrumentId, exposureId=exposureIdStr)
!!*** change this.
!! IMS 10 Jan 2002 - why???
call OAL_setState(instrument=instrumentId, instrExposureId=instExpIdStr)

    call RGSLIB_refreshCache() ! in rgslib/dummycal_mod
    call OAL_exposureInfo(exposureInfo)
    startTime = ToTimeTag(exposureInfo%actualStartTime) ! routine from
    ! caloalutils/caloaldefs

    select case (attitudeStyle) 
    case ('start') 
      if (.not. OAL_getAttitude(startTime, scAttitude)) then
call error('noAttitude', 'No attitude information is available for the &
&start time of the exposure.')
      endif

    case ('mean') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      scAttitude%ra   = real64Attribute(atthkSet,"AAHFRA")  * piOn180
      scAttitude%dec  = real64Attribute(atthkSet,"AAHFDEC") * piOn180
      scAttitude%apos = real64Attribute(atthkSet,"AAHFPA")  * piOn180
      call release(atthkSet)

    case ('median') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      scAttitude%ra   = real64Attribute(atthkSet,"MAHFRA")  * piOn180
      scAttitude%dec  = real64Attribute(atthkSet,"MAHFDEC") * piOn180
      scAttitude%apos = real64Attribute(atthkSet,"MAHFPA")  * piOn180
      call release(atthkSet)

    case ('expmedian')
      attfilterSet = dataSet(attfilterTableName, READ)
      attfilterTab = table(attfilterSet, attfilterTableName)
      scAttitude%ra   = real64Attribute(attfilterTab,"MEDI_RA")  ! These are 
      scAttitude%dec  = real64Attribute(attfilterTab,"MEDI_DEC") ! in radians
      scAttitude%apos = real64Attribute(attfilterTab,"MEDI_APA") ! already
      call release(attfilterSet)

    case default ! note, can't be 'user' because we have dealt with that above.
call error('badAttitudeStyle', 'The value '//trim(attitudeStyle)//' of &
&parameter --attstyle is not recognised.')
    end select

    haveScAttitude = .true.

    ! Correct reference frame for boresight:
    rgsAttitude = instrumentPointing(scAttitude, time=startTime)
    ! routine from rgslib/pointing_mod. Needs CAL_setState hence SAS_CCF.

    call reportRgsAttitude()
  end subroutine getPointing


  !---------------------------------------------------------------------
  subroutine reportRgsAttitude()
  !---------------------------------------------------------------------
  ! Just report (print to STDERR) the Attitude used/determined.

    if (haveScAttitude) then
      call message(AppMsg, Noisy, 'The ScAttitude used now is [in unit of degree]:')
      write(errstr, '(a,f12.7)') " sc_ra   =", scAttitude%ra/piOn180
      call message(AppMsg, Noisy)
      write(errstr, '(a,f12.7)') " sc_dec  =", scAttitude%dec/piOn180
      call message(AppMsg, Noisy)
      write(errstr, '(a,f12.7)') " sc_apos =", scAttitude%apos/piOn180
      call message(AppMsg, Noisy)
    end if

    call message(AppMsg, Verbose, 'The RGS attitude used now is [in unit of degree]:')
    write(errstr, '(a,f12.7)') " attra   =", rgsAttitude%ra/piOn180
    call message(AppMsg, Verbose)
    write(errstr, '(a,f12.7)') " attdec  =", rgsAttitude%dec/piOn180
    call message(AppMsg, Verbose)
    write(errstr, '(a,f12.7)') " attapos =", rgsAttitude%apos/piOn180
    call message(AppMsg, Verbose)

  end subroutine reportRgsAttitude


  !---------------------------------------------------------------------
  subroutine calcMaxNumSrcs(maxNumSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(out) :: maxNumSrcs

    integer(int16) :: numOldListSrcs, numNewEpicSrcs
    type(DataSetT) :: oldListSet, epicSrcSet
    type(TableT) :: oldListTab, epicSrcTab

    !-------------------------------------------------------------------
    ! Start:

    numOldListSrcs = 0
    if (trim(fileMode) .eq. 'modify') then
      ! Open oldListSet and get number of rows:
      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      call release(oldListSet)
    end if

    numNewEpicSrcs = 0
    if (withEpicSet) then
      ! Open epicSrcSet and get number of rows:
      epicSrcSet = dataSet(epicSrcSetName, READ)
      epicSrcTab = table(epicSrcSet, 'SRCLIST', READ)
      numNewEpicSrcs = numberOfRows(epicSrcTab)
      call release(epicSrcSet)
    end if

    maxNumSrcs = numOldListSrcs + numNewEpicSrcs + 2 ! 2 for the proposal and
    ! onaxis sources.

    if (addUserSrc) maxNumSrcs = maxNumSrcs + 1

if (maxNumSrcs .le. 0) call error('bug1', 'Contact the code developer.')

  end subroutine calcMaxNumSrcs

  !---------------------------------------------------------------------
  subroutine fillOutputSrcList(newListSrcData, lastNewListSrc, newEpicAttrNum)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      newListSrcData
    integer(int16), intent(out) ::&
      lastNewListSrc
    integer(int32), intent(out) ::&
      newEpicAttrNum

    integer(int16) ::&
      numSrcs
    type(DataSetT) :: oldListSet, tempEpicSrcSet
    type(TableT) :: oldListTab, tempEpicSrcTab
    type(sourceType), dimension(:), allocatable ::&
      oldListSrcData, newEpicSrcData
    type(sourceType), dimension(1) ::&
      userSrcData
    integer(int16) ::&
      src,&
      src2,&
      numOldListSrcs,&
      maxNumNewEpicSrcs,&
      lastEpicSrc,&
      firstSrcToCheck,&
      previousLastNewListSrc,&
      onAxisNumber,&
      maxPreviousIndex

type(ColumnT) ::&
dispCol,&
xdspCol
real(single), dimension(:), pointer ::&
dispColPtr,&
xdspColPtr
real(double), dimension(:), pointer ::&
raPtr,&
decPtr
type(sourceType) ::&
dummySrcData

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Constructing the source list in memory.')

    numSrcs = size(newListSrcData)

    call initSrcDataStructure(newListSrcData)

    select case(trim(fileMode))
    case('create')
      call addProposalSrc(newListSrcData(1))
      call addOnaxisSrc(  newListSrcData(2))
      lastNewListSrc = 2

      newEpicAttrNum = 1

    case('modify')
!.......................................................................
! Check and load the old source list into memory:

      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      allocate(oldListSrcData(numOldListSrcs))
      call initSrcDataStructure(oldListSrcData)
      call readTable(oldListTab, oldListSrcData)
      call release(oldListSet)

      call checkOldList(oldListSrcData, onAxisNumber)
!********

      ! Load the old list sources into the new array:
      firstSrcToCheck = 1
      lastNewListSrc = 0
      call loadAndTestForLabelClashes(oldListSrcData&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)

      if (changeAttitude) call addOnaxisSrc(newListSrcData(onAxisNumber))

      newEpicAttrNum = 0
      do src = 1, lastNewListSrc
        if (newListSrcData(src)%epicFileAttrNum .ne. myInt32null) then
          newEpicAttrNum = max(newEpicAttrNum&
          , newListSrcData(src)%epicFileAttrNum)
        end if
      end do
      newEpicAttrNum = newEpicAttrNum + 1

    case default
call error('badFileMode', 'The value '//trim(fileMode)//' of parameter &
&--filemode is not recognised.')
    end select ! fileMode.

! At present, we have a newList sources array containing at least the PROPOSAL and ONAXIS at indices 1 and 2, and the remainder of the old list if there is one. So far all the labels of the new list entries are guaranteed to be unique. Next job is to load the new epic sources (if withEpicSet) into a separate array.

    if (withEpicSet) then
!.......................................................................
! Load epic sources into memory:

      tempEpicSrcSet = clone(epicSrcSetName, 'tempEpicSet.ds', TEMP)
!      tempEpicSrcSet = clone(epicSrcSetName, tempSrcSet, TEMP)
!tempEpicSrcSet = clone(epicSrcSetName, tempSrcSet, MODIFY)
!Can use this alternative if it is desired to keep the filtered epic set.

    xPixelSizeWorld = real64Attribute(tempEpicSrcSet, 'TCDELT1')
    yPixelSizeWorld = real64Attribute(tempEpicSrcSet, 'TCDELT2')
    pixelAreaArcsec = abs(xPixelSizeWorld * yPixelSizeWorld&
    * 3600.0 * 3600.0)
!*** can replace by wcs_aux routines?

      tempEpicSrcTab = table(tempEpicSrcSet, 'SRCLIST', MODIFY)
dispCol = addColumn(tempEpicSrcTab, 'DELTA_DISP',   REAL32)
xdspCol = addColumn(tempEpicSrcTab, 'DELTA_XDSP',   REAL32)
dispColPtr => real32Data(dispCol)
xdspColPtr => real32Data(xdspCol)
raPtr => real64Data(column(tempEpicSrcTab, 'RA'))
decPtr => real64Data(column(tempEpicSrcTab, 'DEC'))

do src = 1, numberOfRows(tempEpicSrcTab)
  dummySrcData%raDeg  = raPtr( src - 1)
  dummySrcData%decDeg = decPtr(src - 1)
  call calculateFOVcoords(dummySrcData, rgsAttitude)
  dispColPtr(src - 1) = dummySrcData%deltaDispArcmin
  xdspColPtr(src - 1) = dummySrcData%deltaXdspArcmin
end do

call release(tempEpicSrcTab)
tempEpicSrcTab = table(tempEpicSrcSet, 'SRCLIST', MODIFY)

      maxNumNewEpicSrcs = spatialFilter(tempEpicSrcTab)
      allocate(newEpicSrcData(maxNumNewEpicSrcs))

      call readEpicData(tempEpicSrcTab, newEpicSrcData, lastEpicSrc)
!call deleteColumn(tempEpicSrcTab, 'DELTA_DISP')
!call deleteColumn(tempEpicSrcTab, 'DELTA_XDSP')
      call release(tempEpicSrcSet)

      newEpicSrcData%epicFileAttrNum = newEpicAttrNum

      do src = 1, lastEpicSrc - 1
        do src2 = src + 1, lastEpicSrc
          if (newEpicSrcData(src)%label .eq. newEpicSrcData(src2)%label) then
call error('nonUniqueEpicId', 'Two sources in the supplied EPIC list &
&have been found to have duplicate values of the *_ID_SRC column, resulting &
&in duplication of the label '//trim(newEpicSrcData(src)%label)//'.')
          end if
        end do
      end do
!***** why isn't this test absorbed into loadAndTestForLabelClashes?

      firstSrcToCheck = 1
      previousLastNewListSrc = lastNewListSrc
      call loadAndTestForLabelClashes(newEpicSrcData(1:lastEpicSrc)&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)

      if (lastNewListSrc - previousLastNewListSrc .le. 0) then
        call warning('noValidEpicSources', 'None of the epic sources have &
        &met all the conditions for inclusion.')
      end if
    end if ! withEpicSet

    if (addUserSrc) then
!.......................................................................
! Load user source into memory:
      call getUserSrc(userSrcData(1))

      firstSrcToCheck = 1
      call loadAndTestForLabelClashes(userSrcData&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)
    end if

!.......................................................................
! Now the new or output source list is complete, in that all the sources
! are present. All label values are unique. We have still to fill in or
! correct a few columns though.

    ! ra/dec or fov:
    ! Note: all sources (except all previous rgs sources in 'modify' mode,
    ! and the ONAXIS source in 'create' mode) should have either ra/dec or
    ! xdsp/disp/phi/r set to (my)nulls at this point. After the following loop
    ! none of these 6 columns should have any nulls in them.
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%fixedOnSky) then ! ra and dec should be left
      ! untouched. Do we recalculate the others?
        if (changeAttitude&
        .or. newListSrcData(src)%deltaDispArcmin .eq. myReal32null&
        .or. newListSrcData(src)%deltaXdspArcmin .eq. myReal32null) then ! yes.
          call calculateFOVcoords(newListSrcData(src), rgsAttitude)
        end if

      else ! not 'fixed on sky', the disp/xdisp and fov values should not
      ! be touched. Do we recalculate ra and dec?
        if (changeAttitude&
        .or. newListSrcData(src)%raDeg .eq. myReal32null&
        .or. newListSrcData(src)%decDeg .eq. myReal32null) then ! yes.
          call invertFOVcoords(newListSrcData(src), rgsAttitude)
        end if
      end if
    end do

    ! Calculate INDEX value. First find the maximum non-null existing value:
    maxPreviousIndex = 0
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%index .ne. myInt16null) then
        maxPreviousIndex = max(maxPreviousIndex, newListSrcData(src)%index)
      end if
    end do
    ! Now fill in all the null values:
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%index .eq. myInt16null) then
        maxPreviousIndex = maxPreviousIndex + 1
        newListSrcData(src)%index = maxPreviousIndex
      end if
    end do

! Note: can't do confusion until after prime source selected.

  end subroutine fillOutputSrcList

  !---------------------------------------------------------------------
  subroutine checkOldList(oldListSrcData, onAxisNumber)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(in) ::&
      oldListSrcData
    integer(int16), intent(out) :: onAxisNumber
    logical ::&
      oldListProposalExists,&
      oldListOnaxisExists
    integer(int16) :: src, proposalNumber, numOldListSrcs

    !-------------------------------------------------------------------
    ! Start:

    numOldListSrcs = size(oldListSrcData)

    ! Do some checks of the PROPOSAL source in the old list: 
    oldListProposalExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. proposalLabel) then
        if (oldListProposalExists) then ! there is more than one PROPOSAL
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyProposals', 'There is more than one proposal source &
&in the old RGS source list.')

        else
          oldListProposalExists = .true.
          proposalNumber = src
        end if
      end if
    end do
if (.not. oldListProposalExists) call error('noProposal', 'There is &
&no proposal source in the old RGS source list.')

    ! Do some checks of the ONAXIS source in the old list: 
    oldListOnaxisExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. onAxisLabel) then
        if (oldListOnaxisExists) then ! there is more than one ONAXIS
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyAttSrc', 'There is more than one pointing source &
&in the old RGS source list.')

        else
          oldListOnaxisExists = .true.
          onAxisNumber = src
        end if
      end if
    end do
if (.not. oldListOnaxisExists) call error('noAttSrc', 'There is &
&no pointing source in the old RGS source list.')

!****** other checks on old list??? Check labels are unique, and error if not!!
!****** check matchup between PROCESS/BKG_EXCLUDE and the appropriate region extensions!

  end subroutine checkOldList

  !---------------------------------------------------------------------
  subroutine readEpicData(epicSrcTab, newEpicSrcData, lastEpicSrc)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      epicSrcTab
    type(sourceType), dimension(:), intent(inout) ::&
      newEpicSrcData
    integer(int16), intent(out) ::&
      lastEpicSrc

    type :: rateValsType
      real(single) :: rate, counts, bkgRate
    end type

    integer(int16) ::&
      maxNumNewEpicSrcs, numInsts, i, j&
      , numBands, minBandId, maxBandId
!    integer(int16) ::&
!      src
    real(single), dimension(:), pointer ::&
      epicRatePtr, epicCountsPtr, epicBkgRatePtr
    real(double), dimension(:), pointer ::&
      epicRaPtr, epicDecPtr
    integer(int32), dimension(:), pointer ::&
      epicIndexPtr, epicInstPtr, epicBandPtr, uniqueIdInsts
    integer(int32) :: smallestIndex, largestIndex, indexNum, firstRow
    type(rateValsType), allocatable :: rateData(:,:), rateData_tmp(:,:)	! (SRCID, INST)
    logical, allocatable :: applicableRows(:), applicableRows2(:)
    real(single), dimension(size(newEpicSrcData)) ::&
         denominators, weights_tmp


    !-------------------------------------------------------------------
    ! Start:

    numBands = size(epicBandIds)
    minBandId = minval(epicBandIds)
    maxBandId = maxval(epicBandIds)

    maxNumNewEpicSrcs = size(newEpicSrcData)
    call initSrcDataStructure(newEpicSrcData)

    epicRaPtr => real64Data(column(epicSrcTab, 'RA'))
    epicDecPtr => real64Data(column(epicSrcTab, 'DEC'))
    if (hasColumn(epicSrcTab, 'ML_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'ML_ID_SRC'))
    else if (hasColumn(epicSrcTab, 'BOX_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'BOX_ID_SRC'))
    else
call error('badEpicIdColumn', 'The supplied EPIC source list has neither &
&a ML_ID_SRC nor a BOX_ID_SRC column.')
    end if
    epicInstPtr => int32Data(column(epicSrcTab, 'ID_INST'))
    epicBandPtr => int32Data(column(epicSrcTab, 'ID_BAND'))
    epicRatePtr    => real32Data(column(epicSrcTab, 'RATE'))
    epicCountsPtr  => real32Data(column(epicSrcTab, 'SCTS'))
    epicBkgRatePtr => real32Data(column(epicSrcTab, 'BG_MAP'))

    smallestIndex = minval(epicIndexPtr)
    largestIndex  = maxval(epicIndexPtr)
    uniqueIdInsts => findUniqueValues(epicInstPtr, (epicIndexPtr==smallestIndex .and. epicInstPtr .ne. 0))
    ! in ssclib/array_utils.
    numInsts = size(uniqueIdInsts)
    allocate(rateData(maxNumNewEpicSrcs,numInsts)&
      ,applicableRows(maxNumNewEpicSrcs), applicableRows2(maxNumNewEpicSrcs))
    rateData%rate = 0.0
    rateData%counts = 0.0
    rateData%bkgRate = 0.0

    !! *** Algorithm
    ! All-Columns(:) ==> epicInstPtr(:), epicBandPtr(:), etc.
    ! (Index of Smallest to Largest in All-Columns(:)).Each {
    !   SelectIndex(    current-src, Bands(min-to-max))
    !   raDeg, decDeg, label,... <== firstIndex-of-above
    !   Instruments.Each{
    !     SelectIndex(  current-src, Bands(min-to-max), current-inst)
    !     Bands(2..3).each{
    !       SelectIndex(current-src, current-band,      current-inst)
    !       %rate(src,inst)+=RATE,... <== firstIndex-of-above
    !         ! Input is allowed to have rows with identical Index,Energy,Band.
    !     }
    !   }
    ! }

    lastEpicSrc = 0
!print *, 'smallestIndex, largestIndex', smallestIndex, largestIndex
    do indexNum = smallestIndex, largestIndex
!print *, 'indexNum', indexNum
      applicableRows = (epicIndexPtr==indexNum .and.&
        epicBandPtr>=minBandId .and. epicBandPtr<=maxBandId)
    if (.not.any(applicableRows)) cycle
      lastEpicSrc = lastEpicSrc + 1

      firstRow = smallestTrueIndex(applicableRows) - 1 ! in ssclib/array_utils.
      newEpicSrcData(lastEpicSrc)%raDeg = epicRaPtr(firstRow)
      newEpicSrcData(lastEpicSrc)%decDeg = epicDecPtr(firstRow)
      write(newEpicSrcData(lastEpicSrc)%label, '(a4, i5.5)')&
       trim(epicLabelPrefix), int(indexNum, int16)

      !! Now, calculate rates etc.
!print *, 'firstRow', firstRow
      do i = 1, numInsts	! Each instrument
!print *, 'i, uniqueIdInsts(i)', i, uniqueIdInsts(i)
        if (.not.any(applicableRows .and. (epicInstPtr==uniqueIdInsts(i)))) then
          write(errstr, '(a,i1,a,i3,a)') 'Expecting a value of ID_INST('&
           , i, ') for the ID (', lastEpicSrc&
           , ') in the EPIC source list, but failed to find it.'
          call warning('mangledIdInstColumn')
          cycle
        end if

        applicableRows2 = (applicableRows .and. (epicInstPtr==uniqueIdInsts(i))) ! in ssclib/array_utils.

        do j = 1, numBands	! Each band
          if (.not.any(applicableRows2&
             .and. (epicBandPtr==epicBandIds(j)))) then
            write(errstr, '(a,i1,a,i2,a,i3,a)') 'Expecting a value of ID_INST('&
             , i, ') and of ID_BAND(', j, ') for the ID (', lastEpicSrc&
             , ') in the EPIC source list, but failed to find it.'
            call warning('mangledIdInstBandColumn')
            cycle
          end if

          firstRow = smallestTrueIndex(applicableRows2&
             .and. (epicBandPtr==epicBandIds(j))) - 1 ! in ssclib/array_utils.
          ! firstRow = smallestTrueIndex(applicableRows .and. (epicInstPtr==uniqueIdInsts(i))) - 1 ! in ssclib/array_utils.
          rateData(lastEpicSrc,i)%rate    = rateData(lastEpicSrc,i)%rate &
             + epicRatePtr(   firstRow)	! Simple sum
          rateData(lastEpicSrc,i)%counts  = rateData(lastEpicSrc,i)%counts &
             + epicCountsPtr( firstRow)	! Simple sum
          rateData(lastEpicSrc,i)%bkgRate = rateData(lastEpicSrc,i)%bkgRate &
             + epicBkgRatePtr(firstRow)	! Simple sum
        end do
      end do
    end do

    ! `rateData_tmp' <= same as rateData but with 0.0 for NaN in rateData
    allocate(rateData_tmp(maxNumNewEpicSrcs,numInsts))
    rateData_tmp=rateData
    where (rateData%counts >= 0.0)
    elsewhere 	! namely if NaN
      rateData_tmp%rate    = 0.0
      rateData_tmp%counts  = 0.0
      rateData_tmp%bkgRate = 0.0
    end where

    ! Now calculate weighted averages of the rate-type stuff:

    denominators=0.0	! Array
    weights_tmp=0.0	! Array
    newEpicSrcData%rate    = 0.0
    newEpicSrcData%counts  = 0.0
    newEpicSrcData%bkgRate = 0.0

    do i = 1, numInsts
    if (.not.any(weights%idInst==uniqueIdInsts(i))) cycle

      j = smallestTrueIndex(weights%idInst==uniqueIdInsts(i)) ! in ssclib/array_utils.

      ! `weights_tmp' <= weights%weight but with 0.0 for NaN in weights%weight
      weights_tmp=weights(j)%weight
      where (rateData(:,i)%counts >= 0.0)
      elsewhere 	! namely if NaN
        weights_tmp=0.0
      end where

      denominators = denominators + weights_tmp

      newEpicSrcData%rate    = newEpicSrcData%rate    + rateData_tmp(:,i)%rate
      newEpicSrcData%counts  = newEpicSrcData%counts  + rateData_tmp(:,i)%counts
      newEpicSrcData%bkgRate = newEpicSrcData%bkgRate + rateData_tmp(:,i)%bkgRate
    end do

if (product(denominators)==0.0) call error('badWeightingResult',&
     "Couldn't find weights for any of the ID_INST values in the epic source list.")

    newEpicSrcData%rate    = newEpicSrcData%rate    / denominators
    newEpicSrcData%counts  = newEpicSrcData%counts  / denominators
    newEpicSrcData%bkgRate = newEpicSrcData%bkgRate / denominators

    newEpicSrcData%bkgFlux = newEpicSrcData%bkgRate / pixelAreaArcsec

!**** get newEpicTrueFilename
!newEpicTrueFilename = "don't know!"

    deallocate(uniqueIdInsts)
    deallocate(weights)

  end subroutine readEpicData

  !---------------------------------------------------------------------
  function spatialFilter(epicSrcTab) result(numOfRows)
  !---------------------------------------------------------------------

!    use various_mod

    implicit none 

    type(TableT), intent(inout) ::&
      epicSrcTab
    integer(int16) ::&
      numOfRows, numOfRowsFlagged, numOfAllRows
    type(ColumnT) ::&
      flagCol, flagColOrg1

    !-------------------------------------------------------------------
    ! Start:

    numOfAllRows = numberOfRows(epicSrctab)

    if (enableSpatialFilter) then
      call message(AppMsg, Noisy, "Execute spatial filtering for EPIC sources.")
      if (doFlagEpicSrcOutOfFov) then
        if (hasColumn(epicSrctab, 'FLAG')) then
          if (hasColumn(epicSrctab, 'FLAGORG1')) then
            call error('existFLAGORG1', 'FLAGORG1 already exists.  Contact the developper.')
          else
            flagColOrg1 = column(epicSrctab, 'FLAG')	! Probably "12A"
            call rename(flagColOrg1, 'FLAGORG1')	! Backup
            call release(flagColOrg1)
          end if
        end if
        flagCol  = addColumn(epicSrctab, 'FLAG', INTEGER32, comment = 'Quality flag.')
        call release(flagCol)

        call initFilter('isNull(DELTA_XDSP) || !'//fovFilterStr)	! in dataFilter in selectlib
        numOfRowsFlagged = doSelect(epicSrcTab, 'FLAG', flagBitPos=OUTOFFOV_BIT)	! in dataFilter in selectlib

        write(errStr, '(a,i5,a,i5,a)') "Out of ", numOfAllRows, " rows (EPIC), "&
             , numOfRowsFlagged," rows are flagged outside the FOV."
        call message(AppMsg, Verbose)
        numOfRows = numOfAllRows	! Number of rows that exist in Table

      else
        call message(AppMsg, Verbose, 'Drop EPIC sources outside the FOV.')
        call initFilter(fovFilterStr)		! in dataFilter in selectlib
        numOfRows = doFilter(epicSrcTab)	! in dataFilter in selectlib
        numOfRowsFlagged = numOfAllRows - numOfRows

        write(errStr, '(a,i5,a,i5,a)') "Out of ", numOfAllRows, " rows (EPIC), "&
             , numOfRowsFlagged&
             , " rows are dropped due to their being outside the FOV."
        call message(AppMsg, Verbose)

      end if

      if (numOfRows .le. 0) then
        call warning('noEpicSourcesInFov', 'None of the epic sources &
        &were found to lie within the field of view of the RGS camera.')
      end if

    else
      numOfRows = numOfAllRows
      if (numOfRows .le. 0) then
        call warning('noEpicSources', 'There are no epic sources in &
        &the correct energy band.')
      end if
    end if

  end function spatialFilter

  !---------------------------------------------------------------------
  subroutine getUserSrc(userSrcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(out) ::&
      userSrcData

    !-------------------------------------------------------------------
    ! Start:

    if (userSrcLabel .eq. '') then
call error('emptyUserLabel', 'The label for the user-specified source is &
&not permitted to consist only of blanks.')
    end if

!!***** NOTE!! The following checks are designed mostly to detect when the user has allowed the ra, dec, deltadisp etc parameters to take their default values. The user should not do this. The checks depend to some degree on the values of the relevant parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
!    if (userSrcStyle .eq. 'radec') then
!      if (userSrcRaDeg .lt. 0.0 .or. userSrcRaDeg .gt. 360.0) then
!call error('illegalUserRa', 'The --ra for the user-specified source is either &
!&< 0  or > 360 degrees. Have you forgotten to set it?')
!      end if
!      if (userSrcDecDeg .lt. -90.0 .or. userSrcDecDeg .gt. 90.0) then
!call error('illegalUserDec', 'The --dec for the user-specified source is &
!&either < -90 or > 90 degrees. Have you forgotten to set it?')
!      end if
!
!    else ! userSrcStyle .eq. 'wrtatt'
!      if (userSrcXdspArcmin .eq. 0.0 .or. userSrcDispArcmin .eq. 0.0) then
!        call warning('zeroWrtAttCoord', 'Either --deltaxdsp or --deltadisp &
!        &has been left at its default value of 0 arcminutes. Have you &
!        &forgotten to set it?')
!      end if
!    end if

    call initSrcDataStructure(userSrcData)

!    userSrcLabel = no0s(userSrcLabel)
    userSrcData%label = userSrcLabel
    if (userSrcStyle .eq. 'radec') then
      userSrcData%raDeg  = userSrcRaDeg
      userSrcData%decDeg = userSrcDecDeg
      userSrcData%fixedOnSky = .true.

    else ! userSrcStyle .eq. 'wrtatt'
      userSrcData%deltaDispArcmin = userSrcDispArcmin
      userSrcData%deltaXdspArcmin = userSrcXdspArcmin
      call dispXdspToPolar(&
        userSrcData%deltaDispArcmin,&
        userSrcData%deltaXdspArcmin,&
        userSrcData%fovPhiDeg,&
        userSrcData%fovRdeg&
      )
      userSrcData%fixedOnSky = .false.
    end if
    userSrcData%process = processUserSrc
    userSrcData%bkgExclude = bkgExcludeUserSrc
    userSrcData%rate = userSrcRate

  end subroutine getUserSrc

  !---------------------------------------------------------------------
  subroutine readTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(out) ::&
      srcData

    integer(int16) ::&
      numSrcs
    integer(int32) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(kind=bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)

    ! write column values to data array:
    srcData(1:numSrcs)%index           = indexColPtr
    srcData(1:numSrcs)%raDeg           = raColPtr
    srcData(1:numSrcs)%decDeg          = decColPtr
    srcData(1:numSrcs)%rate            = rateColPtr
    srcData(1:numSrcs)%deltaDispArcmin = dispColPtr
    srcData(1:numSrcs)%deltaXdspArcmin = xdspColPtr
    srcData(1:numSrcs)%fovPhiDeg       = phiColPtr
    srcData(1:numSrcs)%fovRdeg         = rColPtr
    srcData(1:numSrcs)%confusion       = confnColPtr
    srcData(1:numSrcs)%process         = procColPtr
    srcData(1:numSrcs)%bkgExclude      = excluColPtr
    srcData(1:numSrcs)%fixedOnSky      = fixedColPtr

    do src = 1, numSrcs
      if (isNull(epicnCol, src - 1)) then
        srcData(src)%epicFileAttrNum = myInt32Null
      else
        srcData(src)%epicFileAttrNum = epicnColPtr(src - 1)
      end if

      srcData(src)%label = removeTrailingChar0s(trim(stringCell(labelCol&
        , src - 1)))
    end do

  end subroutine readTable

  !---------------------------------------------------------------------
  subroutine addProposalSrc(srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(inout) ::&
      srcData

    type(ProposalInfoType) :: proposalInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Finding proposal info from the ODF.')

    call OAL_proposalInfo(proposalInfo)
!**** presumably the state of the OAL doesn't have to be set first? It isn't guaranteed to be set at the moment anyway... eg if attstyle=user.
    srcData%raDeg  = proposalInfo%targetRa
    srcData%decDeg = proposalInfo%targetDec
    srcData%label  = proposalLabel 

    ! All others: leave as defaults.

  end subroutine addProposalSrc

  !---------------------------------------------------------------------
  subroutine addOnaxisSrc(srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(inout) ::&
      srcData

    !-------------------------------------------------------------------
    ! Start:

    srcData%deltaDispArcmin = 0.0
    srcData%deltaXdspArcmin = 0.0
    call dispXdspToPolar(&
      srcData%deltaDispArcmin,&
      srcData%deltaXdspArcmin,&
      srcData%fovPhiDeg,&
      srcData%fovRdeg&
    )
    srcData%label = onAxisLabel 
    srcData%fixedOnSky = .false.
    srcData%raDeg  = rgsAttitude%ra  / piOn180
    srcData%decDeg = rgsAttitude%dec / piOn180

    ! All others: leave as defaults.

  end subroutine addOnaxisSrc

  !---------------------------------------------------------------------
  subroutine loadAndTestForLabelClashes(trialData, firstTrialSrc&
  , firmData, lastFirmSrc)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      firstTrialSrc
    integer(int16), intent(inout) ::&
      lastFirmSrc
    type(sourceType), dimension(:), intent(in) ::&
      trialData
    type(sourceType), dimension(:), intent(inout) ::&
      firmData
    logical :: labelFoundBefore
    integer(int16) :: trialSrc, firmSrc

    !-------------------------------------------------------------------
    ! Start:

    do trialSrc = firstTrialSrc, size(trialData)
      labelFoundBefore = .false.
      INNER: do firmSrc = 1, lastFirmSrc
        if (trialData(trialSrc)%label .eq. firmData(firmSrc)%label) then
          labelFoundBefore = .true.
          write(errstr, '(a)') 'Duplicate label: '&
          //trim(trialData(trialSrc)%label)
          if (clobberOnLabelClash) then
            call warning('duplicateLabel')
      exit INNER

          else
call error('duplicateLabel')
          end if
        end if
      end do INNER

      if (.not. labelFoundBefore) then
        lastFirmSrc = lastFirmSrc + 1
if (lastFirmSrc .gt. size(firmData)) call error('bug2', 'Contact the code developer.')
        firmData(lastFirmSrc) = trialData(trialSrc)
      end if
    end do

!***** if user source is one that clashes, should overwrite the other source, not delete the user source??? probably not.

!***** this thing should return statuses, and let the calling routine do the error/warnings????

  end subroutine loadAndTestForLabelClashes

  !---------------------------------------------------------------------
  subroutine writeToDataSet(srcData, instrumentName, newEpicAttrNum&
  , instExpIdStr, set, tab)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      srcData
    character(len=*) :: instrumentName
    integer(int32), intent(in) :: newEpicAttrNum
    character(*), intent(in) :: instExpIdStr
    type(DataSetT), intent(out) :: set
    type(TableT), intent(out) :: tab

    integer(int16) ::&
      numSrcs

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing the source list to the outset.')

    numSrcs = size(srcData)

    select case(trim(fileMode))
    case('create')
      ! Open file for write:
      set = dataSet(newListSetName, CREATE)

      call writeHeaderKeywords(set, trim(instrumentName), instExpIdStr)

      tab = addTable(set, 'SRCLIST', int(numSrcs))

      if (withEpicSet) then
        call writeEpicSpecificKwds(tab, newEpicAttrNum)
      end if

      call writeNewTable(tab, srcData)

    case('modify')
      ! Open file for modify:
      set = dataSet(oldListSetName, MODIFY)

      if (changeAttitude) then
        call deleteRegionExtensions(set)
        srcData%process = .false.
        srcData%bkgExclude = .false.

        call setAttribute(set, 'RA_PNT',  rgsAttitude%ra   / piOn180)
        call setAttribute(set, 'DEC_PNT', rgsAttitude%dec  / piOn180)
        call setAttribute(set, 'PA_PNT',  rgsAttitude%apos / piOn180)
      end if

      tab = table(set, 'SRCLIST')

      if (withEpicSet) then
        call writeEpicSpecificKwds(tab, newEpicAttrNum)
      end if

      call deleteRows(tab, 0, numberOfRows(tab) - 1)
      call insertRows(tab, 0, int(numSrcs) - 1)
      call writeOldTable(tab, srcData)

    case default
call error('badFileMode', 'The value '//trim(fileMode)//' of parameter &
&--filemode is not recognised.')
    end select

    ! Leaves set, tab open on return.

  end subroutine writeToDataSet

  !---------------------------------------------------------------------
  subroutine reportOofovFlags(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) :: tab
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    integer(int16) ::&
      src, numSrcs
    type(ColumnT) ::&
      flagCol
    integer(int32), dimension(:), pointer ::&
      flagColPtr
character(10) :: indexStr
!integer :: status

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    flagCol  = column(tab, 'FLAG')
    flagColPtr  => int32Data( flagCol)
    srcData%flag = flagColPtr
    call release(flagCol)

    ! Check if any srcData%flags have been set to OUTOFFOV_MASK and warn if so.
    do src = 1, numSrcs
!      if (srcData(src)%flag .eq. 1) then
if (btest(srcData(src)%flag, OUTOFFOV_BIT)) then
!call intToStr(int(src), 'left', indexStr, status)
indexStr = intToStr(int(src), leftJustify=.true._bool)
!        write(errstr, '(a, i5, a)') 'Source INDEX=', src&
write(errstr, '(a, i5, a)') 'Source INDEX='//trim(indexStr)&
//' ('//trim(srcData(src)%label)//') is outside the field of &
        &view of the RGS camera.'
!***** may need to alter this when and if flag column comes into being in the output source list. In that case this loop should also be outside the 'if(enable etc)' in the calling level.
        call warning('outsideFOV')
      end if
    end do

  end subroutine reportOofovFlags

  !---------------------------------------------------------------------
  subroutine getEpicAttData(epicInstrumentId, epicTimeStamp)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(out) :: epicInstrumentId
    real(double),   intent(out) :: epicTimeStamp

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(epicSrcSetName, READ)
    epicInstrumentId = instId(stringAttribute(set, 'INSTRUME'))
    epicTimeStamp = toTimeTag(stringAttribute(set, 'DATE-OBS'))

    if (.not. haveScAttitude) then
!      call warning('', '')

      scAttitude%ra   = real64Attribute(set, "RA_PNT")  * piOn180
      scAttitude%dec  = real64Attribute(set, "DEC_PNT") * piOn180
      scAttitude%apos = real64Attribute(set, "PA_PNT")  * piOn180
!***** NOTE!!! these kwds OUGHT to contain the epic BORESIGHT-CORRECTED attitude. However they seem at the moment to contain the median values from the atthkgen file.
    end if

    call release(set)

  end subroutine getEpicAttData

  !---------------------------------------------------------------------
  subroutine calcEpicPlaneConfusion(epicInstrumentId, epicTimeStamp, srcData)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(in) ::&
      epicInstrumentId
    real(double),   intent(in) ::&
      epicTimeStamp
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    type(srcEnvelopeType) ::&
      subsetSrcData(size(srcData))
    integer(int16) ::&
      i,&
      j,&
      numSrcs,&
      numSubsetSrcs
    real(double),  dimension(:), allocatable ::&
      subsetSrcXdet,&
      subsetSrcYdet,&
      subsetSrcRadiusDet
!    real(single),  dimension(:), allocatable ::&
real(double),  dimension(:), allocatable ::&
      srcXdet,&
      srcYdet,&
      srcFovRadiusDet
    integer(int16),  dimension(:), allocatable ::&
      subsetConfSetNum
    logical(bool) :: srcIsValid(size(srcData))

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    allocate(srcXdet(numSrcs), srcYdet(numSrcs), subsetSrcXdet(numSrcs)&
      , subsetSrcYdet(numSrcs), srcFovRadiusDet(numSrcs))

    call CAL_setState(instrument=epicInstrumentId)
!    call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
!      , epicTimeStamp, srcXdet, srcYdet)
call raDecToDet(srcData%raDeg*PiOn180, srcData%decDeg*PiOn180, scAttitude&
 , epicTimeStamp, srcXdet, srcYdet) ! in ssclib/coordinate.

    srcFovRadiusDet = sqrt(srcXdet * srcXdet + srcYdet * srcYdet)

    ! Get subset of those srcs with flag = 0
    subsetSrcData%EeV = (bandMinEeV + bandMaxEeV) / 2.0
    numSubsetSrcs = 0
    do i = 1, numSrcs
      if (srcData(i)%flag .eq. 0 .and. srcData(i)%label .ne. 'ONAXIS') then
        numSubsetSrcs = numSubsetSrcs + 1
        subsetSrcData(numSubsetSrcs)%bkgFlux = srcData(i)%bkgFlux

        subsetSrcData(numSubsetSrcs)%thetaArcsec&
          = srcFovRadiusDet(i) * detUnitArcsec
!***** shouldn't you use detToTel here??

        subsetSrcData(numSubsetSrcs)%peakFlux = srcData(i)%counts&
          / getPsfIntegral(subsetSrcData(numSubsetSrcs)%thetaArcsec&
          , subsetSrcData(numSubsetSrcs)%EeV)
        ! in counts per square arc second

        subsetSrcXdet(numSubsetSrcs) = srcXdet(i)
        subsetSrcYdet(numSubsetSrcs) = srcYdet(i)
      end if
    end do

    allocate(subsetSrcRadiusDet(numSubsetSrcs)&
      , subsetConfSetNum(numSubsetSrcs))

    call getPsfEllipseAxes(subsetSrcData%thetaArcsec, subsetSrcData%EeV&
      , subsetSrcData%peakFlux, subsetSrcData%bkgFlux, fracBackground&
      , subsetSrcData%longAxisArcsec, subsetSrcData%shortAxisArcsec, srcIsValid)

    ! Take long axes as src radii
    subsetSrcRadiusDet = subsetSrcData(1:numSubsetSrcs)%longAxisArcsec&
    / detUnitArcsec
!**** Simpler to start with to try:   subsetSrcRadiusDet = 5.0 / detUnitArcsec

    call findConfusedSets(subsetSrcXdet(1:numSubsetSrcs)&
      , subsetSrcYdet(1:numSubsetSrcs), subsetSrcRadiusDet(1:numSubsetSrcs)&
      , subsetConfSetNum)!, status)

    ! Write srcData%confusionSet values
    srcData%confusionSet = 0 ! default (indicates no confusion)
    j = 0
    do i = 1, numSrcs
      if (srcData(i)%flag == 0 .and. srcData(i)%label .ne. 'ONAXIS') then
        j = j + 1
        srcData(i)%confusionSet = subsetConfSetNum(j)
      end if
    end do

  end subroutine calcEpicPlaneConfusion

  !---------------------------------------------------------------------
  subroutine flagConfusedSources(tab, srcData)
  !---------------------------------------------------------------------
  ! The basic idea here is to look at each confused set and flag all but
  ! the brightest. However the procedure is slightly different in the
  ! case that the proposal src is a member of a confused set. The most
  ! common physical situation in which this can occur is if the source
  ! is bright enough to cause pileup in epic. In this case the centre of
  ! the source (where the proposal position is) will be burnt out and the
  ! epic source detection chain will detect multiple sources in a ring
  ! around the true position, instead of just a single source approximately
  ! aligned with the proposal. In this pileup case, hopefully all the
  ! sources (ie false epic detections as well as the proposal) will be
  ! confused. In this case the proposal position is best since rgs is
  ! not subject to significant pileup.
  !
  ! However the above case needs to be distinguished from the case of a
  ! source which is not piled up. In this case the danger to watch out
  ! for is the proposal source position being slightly incorrect (this
  ! can easily happen, since it is entered by the observer). So in the
  ! case that the proposal is confused with just one other epic source, the
  ! epic source is chosen instead for non-flagging.
  !
  ! If the source is piled-up but the proposal position incorrect, then
  ! the rgs has no way to find the true centre-source position.

    implicit none 

    type(TableT), intent(in) :: tab
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    type(ColumnT) :: flagCol
    integer(int32), pointer :: flagColPtr(:)
    integer(int16) ::&
      numSrcs, maxConfusionIndex, confSetNum, numSourcesInSet, brightestIndex(1)

    !-------------------------------------------------------------------
    ! Start:

    flagCol  = column(tab, 'FLAG')
    flagColPtr  => int32Data(flagCol)
    srcData%flag = flagColPtr

    numSrcs = size(srcData)
    if (numSrcs .ne. numberOfRows(tab))&
call error('bug10', 'Should never happen - contact the code developer.')

    maxConfusionIndex = maxval(srcData%confusionSet)
    if (maxConfusionIndex .le. 0) then
      call message(AppMsg, Verbose, 'No confused epic sources detected.')
  return
    end if

    do confSetNum = 1, maxConfusionIndex
      numSourcesInSet = count(srcData%confusionSet .eq. confSetNum&
      .and. srcData%flag .eq. 0) ! test for flag is redundant but
      ! let's play safe.

if (numSourcesInSet .eq. 1) call error('bug5', 'Contact the code developer.')

    if (numSourcesInSet .le. 0) cycle ! Shouldn't happen with present
    ! structure of findConfusedSets() but let's play safe.

      if (any(srcData%label .eq. proposalLabel .and. srcData%flag .eq. 0&
      .and. srcData%confusionSet .eq. confSetNum)) then
        if (numSourcesInSet .eq. 2) then ! a nice clean detection. Flag the
        ! proposal and leave the epic position unflagged:

          where(srcData%label .eq. proposalLabel)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)

        else ! probably piled-up or otherwise with messy psf and thus
        ! accompanied by a cluster of false detections. Flag all _but_
        ! the proposal:

          where(srcData%label .ne. proposalLabel&
          .and. srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)
        end if

      else ! PROPOSAL is not a member of the confused set. Flag all but
      ! the brightest.
        brightestIndex = maxloc(srcData%rate, mask=(srcData%flag .eq. 0&
        .and. srcData%confusionSet .eq. confSetNum))

        if (srcData(brightestIndex(1))%rate .le. 0.0) then
          call warning('confSetAllZeroRate', 'A set of confused sources &
          &has none with a positive value of RATE. All have been flagged.')

          where(srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)

        else ! Flag all but the brightest.
          where(srcData%index .ne. srcData(brightestIndex(1))%index&
          .and. srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)
        end if
      end if
    end do

    flagColPtr = srcData%flag 
    call release(flagCol)

  end subroutine flagConfusedSources

  !---------------------------------------------------------------------
  function calcPrimeIndex(oldListTab, srcData) result(indexOfPrime)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(in) ::&
      srcData
    type(TableT), intent(in) :: oldListTab

    integer(int16) ::&
      numSrcs
    integer(int16) ::&
      indexOfPrime, oldIndexOfPrime, proposalIndex, pointingIndex, src,&
      dummy(1), indexOfBrightest
    real(single) :: valueOfBrightest
    logical :: notFound

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating the INDEX value of the &
    &prime source.')

    ! NOTE!!! The algorithm in this subroutine relies for some extent on the parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.

    numSrcs = size(srcData)

!.......................................................................
! Get some index values which might be useful:

    ! Get the previous prime source number if there is one:
    if (trim(fileMode) .eq. 'modify') then
      oldindexOfPrime = int16Attribute(oldListTab, 'PRIMESRC')
    end if

    ! Get the index of the proposal source (better not assume it is 1):
    notFound = .true. ! default
    do src = 1, numSrcs
      if (srcData(src)%label .eq. proposalLabel) then
        notFound = .false.
        proposalIndex = srcData(src)%index
    exit
      end if
    end do

if (notFound) call error('bug6', 'No proposal source found. Contact the code developer.')

    ! Get the index of the pointing source (better not assume it is 2):
    notFound = .true. ! default
    do src = 1, numSrcs
      if (srcData(src)%label .eq. onAxisLabel) then
        notFound = .false.
        pointingIndex = srcData(src)%index
    exit
      end if
    end do

if (notFound) call error('bug7', 'No pointing source found. Contact the code developer.')

!.......................................................................
! Start of main prime source choice. Mostly the next section is taken up by
! the following: (fileMode .eq. 'create' .or. changePrimeSrc) .and. addUserSrc
! .and. setUserSrcToPrime.

indexOfPrime = -1 ! an impossible default, to pick up bugs in this routine. Eventually this line (and the tests at the end of the sub) could probably be dispensed with.

    if (trim(fileMode) .eq. 'modify' .and. .not. changePrimeSrc) then
      indexOfPrime = oldindexOfPrime

    else ! either fileMode .eq. 'create' .or. changePrimeSrc
      if (addUserSrc .and. setUserSrcToPrime) then
        ! Find user source:
        notFound = .true. ! default
        do src = 1, numSrcs
          if (srcData(src)%label .eq. userSrcLabel) then
            notFound = .false.
            indexOfPrime = srcData(src)%index
        exit
          end if
        end do
if (notFound) call error('bug8', 'Contact the code developer.') ! in current implementation user _label_ should always be present, even if there was a label clash with another src and the _actual_ user source has been discarded. In this case indexOfPrime is _not_ going to point to the user source but the existing source that had the same label value. Hence in no case should this error occur.

!.......................................................................
! Remainder of the choice tree deals with non-user prime source cases.

      else ! either no user source supplied, or it wasn't requested to be
      ! the prime source. Look now at the dedicated prime source parameters.
        ! Calculate indexOfPrime:
        select case(primeSrcStyle)
        case('label')
!**** check for all-blank primeSrcLabel? Remember that all-blank labels are not allowed. Same with epic label prefix???
          notFound = .true.
          do src = 1, numSrcs
            if (srcData(src)%label .eq. primeSrcLabel) then
              notFound = .false.
              indexOfPrime = srcData(src)%index
          exit
            end if
          end do

          if (notFound) then
            indexOfPrime = proposalIndex
            call warning('primeLabelNotFound', 'Label '//trim(primeSrcLabel)&
            //' not found; prime source set to proposal source.')
          end if

        case('index')
          notFound = .true.
          do src = 1, numSrcs
            if (srcData(src)%index .eq. paramIndexOfPrime) then
              notFound = .false.
              indexOfPrime = srcData(src)%index
          exit
            end if
          end do

          if (notFound) then
            indexOfPrime = proposalIndex
            write(errstr, '(a, i5, a)') 'Index ', paramIndexOfPrime&
            , ' not found; prime source set to proposal source.'
            call warning('primeIndexNotFound')
          end if

        case('expression')
call error('notSupported', 'Sorry, this mode of entering the prime source &
&is not yet supported.')

        case('brightest')
          dummy = maxloc(srcData%rate)
!***** mask it by flag > 0??? Should also consider impact of flag column on other prime source cases - eg generate some warnings if the prime source is flagged?
          indexOfPrime = dummy(1)
          if (srcData(indexOfPrime)%rate .le. 0.0) then
            indexOfPrime = proposalIndex
            call warning('noBrightPrime', 'The brightest source has a rate &
            &less than or equal to zero; prime source set to proposal source.')
          end if

        case('auto') ! to be used by the pcms ('auto' should probably also
        ! be the default value of this parameter). In this case the code
        ! makes every possible attempt to ensure that the most sensible
        ! possible source is chosen to be the prime src.

          if (srcData(proposalIndex)%flag .eq. 0) then ! ie, if proposal src
          ! is neither confused nor outside the fov:

            indexOfPrime = proposalIndex

          else ! choose the brightest epic source:
            indexOfBrightest = -1 ! nonsense default.
            valueOfBrightest = 0.0
            do src = 1, numSrcs
              if (  srcData(src)%epicFileAttrNum .ne. myInt32null&
              .and. srcData(src)%flag            .eq. 0&
              .and. srcData(src)%rate            .gt. valueOfBrightest) then
                indexOfBrightest = src
                valueOfBrightest = srcData(src)%rate
              end if
!*** can do this via masked maxloc?
            end do
            if (valueOfBrightest .gt. 0.0) then
              indexOfPrime = indexOfBrightest

            else ! if all else fails, choose the pointing direction!
              indexOfPrime = pointingIndex
            end if
          end if

        case default
call error('badPrimeStyle', 'The value of parameter --primestyle is &
&not recognized.')
        end select ! primeSrcStyle
      end if ! addUserSrc .and. setUserSrcToPrime
    end if ! fileMode .eq. 'modify' .and. .not. changePrimeSrc

!.......................................................................
! Finally, do some checks on defaults etc.

    if (addUserSrc .and. setUserSrcToPrime .and. .not. changePrimeSrc) then
      call warning('noChangePrime', 'You have requested that the user &
      &source be the prime source, but have not enabled a change in &
      &prime source via the parameter --changeprime. Prime source not &
      &changed.')
    end if

    if (fileMode .eq. 'modify' .and. changePrimeSrc) then
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        write(errstr, '(a, i5, a)') 'You have requested that the &
        &prime source be changed via the parameter --changeprime, but you &
        &may have forgotten to specify the new prime source, because &
        &it is still set to its default.'

        call warning('primeAtLabelDefault')
      end if

      if (oldindexOfPrime .eq. indexOfPrime) then
        write(errstr, '(a, i5, a)') 'You have requested that the &
        &prime source be changed via the parameter --changeprime, but you &
        &may have forgotten to specify the new prime source index, because &
        &the new prime source index of ', indexOfPrime, ' is the same as &
        &the old one.'

        call warning('primeAtIndexDefault')
      end if

    else if (fileMode .eq. 'create'&
    .and. .not. (addUserSrc .and. setUserSrcToPrime)) then
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        write(errstr, '(a, i5, a)') 'Just a reminder that you may want &
        &to specify the prime source - it is still set to its default.'

        call warning('primeAtDefault')
      end if
    end if

    if (btest(srcData(indexOfPrime)%flag, OUTOFFOV_BIT)) then
      call warning('primeOutsideFOV', 'The prime source is outside the &
      &field of view of the RGS camera. This will have a deleterious &
      &effect on the attitude drift correction of the RGS events.')
!***** may need to alter this when and if flag column comes into being in the output source list.
    end if

if (indexOfPrime .le. 0) call error('bug9', 'Contact the code developer.') ! to pick up bugs. Eventually this test (and the default-setting line) could probably be dispensed with.

  end function calcPrimeIndex

  !---------------------------------------------------------------------
   subroutine writeHeaderKeywords(set, instrumentName, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    character(len=*), intent(in) :: instrumentName
    character(*), intent(in) :: instExpIdStr

    integer(int32) :: instrumentId

    !-------------------------------------------------------------------
    ! Start:

    call writeGenericAttributes(set, instrumentName, rgsAttitude)
    ! 'TELESCOP'
    ! 'INSTRUME'
    ! 'FILTER'
    ! 'ORIGIN'
    ! 'RA_PNT'
    ! 'DEC_PNT'
    ! 'PA_PNT'
    ! 'RADECSYS'
    ! 'EQUINOX'
!    call setAttribute(set, 'EXPOSURE',)
!    call setAttribute(set, 'DATE-OBS',)
!    call setAttribute(set, 'DATE-END',)
!    call setAttribute(set, 'OCB',) ! special RGS header kwd: on-chip binning
!***** ...but the latter is ccd-specific. Tell Julian it should be dropped from the icd-specified list for rgs source lists.

    if (writeObsAttributes) call writeObservationAttributes(set)
    ! 'OBS_MODE'
    ! 'OBS_ID'
    ! 'OBJECT'
    ! 'OBSERVER'

    if (writeExpAttributes) then
      instrumentId = instId(instrumentName)
      call writeExposureAttributes(set, instrumentId, instExpIdStr)
    ! 'DATAMODE'
    ! 'EXP_ID'
    ! 'EXPIDSTR'
    end if

  end subroutine writeHeaderKeywords

  !---------------------------------------------------------------------
  subroutine writeEpicSpecificKwds(outTab, newEpicAttrNum)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in)   :: outTab
    integer(int32), intent(in) :: newEpicAttrNum

    type(DataSetT) :: epicSrcSet
    character(len=8) ::&
      epicExpIdAttrName,&
      epicContentAttrName,&
      epicBandAttrName,&
      epicFilterAttrName,&
      epicExpId,&
      epicFilter
!**** why call it expId when it gets the instrument name?
!**** is 8 char enough for the filter?
    character(len=80) ::&
      epicContent
    integer(int16) :: i, sizeEpicBandIds, sizeEpicBandIds4Key = 9
    integer(int32) ::   newEpicAttrNum2
    character(len=2) :: newEpicAttrNum2char

    !-------------------------------------------------------------------
    ! Start:

    !! Check for newEpicAttrNum over 99 
    newEpicAttrNum2 = newEpicAttrNum
    if (newEpicAttrNum > 99) then
      newEpicAttrNum2 = mod(newEpicAttrNum, 100)
      write(newEpicAttrNum2char, *) newEpicAttrNum2 
      write(errstr, '(a, i3, a, i4, a)') 'Too many (',newEpicAttrNum&
       , ') EPIC source lists are input.  The header keywords of E_*'&
       //trim(newEpicAttrNum2char)//' are used (maybe overwritten) for the sequence '&
       , newEpicAttrNum, '.'
      call warning('tooBigEPICAttrNum')
    end if

    !! Read some keywords
    epicSrcSet = dataSet(epicSrcSetName, READ)
      epicExpId   = stringAttribute(epicSrcSet, 'INSTRUME')
      epicContent = stringAttribute(epicSrcSet, 'CONTENT')
      epicFilter  = stringAttribute(epicSrcSet, 'FILTER')
    call release(epicSrcSet)

    write(epicExpIdAttrName,   '(a6, i2)') 'E_EXPR', newEpicAttrNum2
    write(epicContentAttrName, '(a6, i2)') 'E_CONT', newEpicAttrNum2
    write(epicFilterAttrName,  '(a6, i2)') 'E_FILT', newEpicAttrNum2
    call setAttribute(outTab, epicExpIdAttrName,   epicExpId,   COMMEPICEXPID)
    call setAttribute(outTab, epicContentAttrName, epicContent, COMMEPICCONTENT)
    call setAttribute(outTab, epicFilterAttrName,  epicFilter,  COMMEPICFILTER)

    !! Writing E_mBNDn
    ! write(epicBandAttrName,    '(a6, i2)') 'E_BAND', newEpicAttrNum2
    sizeEpicBandIds = size(epicBandIds)
    if (sizeEpicBandIds > 9) then
       ! sizeEpicBandIds4Key = 9	! Default
       ! call warning('tooManyEpicBandIds') in readParameters
    else
       sizeEpicBandIds4Key = sizeEpicBandIds
    end if

    do i=1, sizeEpicBandIds4Key
      if (newEpicAttrNum2 > 9) then		! if 2 digits
        write(epicBandAttrName,  '(a2,i1,a3,i2)') 'E_',i,'BND',newEpicAttrNum2
      else
        write(epicBandAttrName,  '(a2,i1,a3,i1)') 'E_',i,'BND',newEpicAttrNum2
      end if

      !! The following check should not be needed, unless 'tooBigEPICAttrNum';
      !  this must not happen when \param{filemode}!=`modify' first,
      !  and even when \param{filemode}=`modify', the sequence-ID
      !  `newEpicAttrNum' in the input source list must differ from
      !  the current one, which is being used here.
      if (hasAttribute(outTab, epicBandAttrName)) then
        write(errstr, '(a, a, a)') 'Keyword ', epicBandAttrName&
             , ' already exists.  Updated in the output FITS.'
        call warning('keywordBNDUpdated')
      end if

      call setAttribute(outTab, epicBandAttrName, epicBandIds(i), COMMEPICBANDNUM)
    end do

    !! Delete pre-existing E_iBNDnn keywords (for n==newEpicAttrNum)
    !  This check should not be needed, unless 'tooBigEPICAttrNum' (see above).
    do i=sizeEpicBandIds4Key+1, 9	! 9 is the maximum number due to 8-char limit.
      if (newEpicAttrNum2 > 9) then		! if 2 digits
        write(epicBandAttrName,  '(a2,i1,a3,i2)') 'E_',i,'BND',newEpicAttrNum2
      else
        write(epicBandAttrName,  '(a2,i1,a3,i1)') 'E_',i,'BND',newEpicAttrNum2
      end if

      if (hasAttribute(outTab, epicBandAttrName)) then
        write(errstr, '(a, a, a)') 'Keyword ', epicBandAttrName&
             , ' already exists.  Deleted in the output FITS.'
        call warning('keywordBNDDeleted')
        call deleteAttribute(outTab, epicBandAttrName)
      end if
    end do

  end subroutine writeEpicSpecificKwds

  !---------------------------------------------------------------------
  subroutine writeOldTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(in) ::&
      srcData

    integer(int16) ::&
      numSrcs, src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol,&
      flagCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr,&
      flagColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')
if (hasColumn(tab, 'FLAG')) then
flagCol  = column(tab, 'FLAG')
else
flagCol  = addColumn(tab, 'FLAG', INTEGER32, comment = 'Quality flag.')
!***** because there isn't one yet in the disk file format.
end if

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)
    flagColPtr  => int32Data( flagCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', 'Could not set an int32 null value for &
&column EPIC_FILE.')
    end if

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate
      dispColPtr(src - 1)  = srcData(src)%deltaDispArcmin
      xdspColPtr(src - 1)  = srcData(src)%deltaXdspArcmin
      phiColPtr(src - 1)   = srcData(src)%fovPhiDeg
      rColPtr(src - 1)     = srcData(src)%fovRdeg
      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky
      flagColPtr(src - 1)  = srcData(src)%flag

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

      call setStringCell(labelCol , src - 1, srcData(src)%label)
! DON'T trim the string argument to setStringCell. If you do, the characters
! in the file that should be trailing blanks will not be overwritten from
! whatever they were before in the memory map of the file, and will thus
! contain nonsense characters.
    end do

    call release(indexCol)
    call release(labelCol)
    call release(raCol)
    call release(decCol)
    call release(rateCol)
    call release(dispCol)
    call release(xdspCol)
    call release(phiCol)
    call release(rCol)
    call release(confnCol)
    call release(procCol)
    call release(excluCol)
    call release(fixedCol)
    call release(epicnCol)
    call release(flagCol)

  end subroutine writeOldTable

  !---------------------------------------------------------------------
  subroutine deleteRegionExtensions(set)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    integer :: blockNumber

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Deleting all region extensions.')

    do blockNumber = 1, numberOfBlocks(set) - 1 ! start from 1 not 0 because
    ! we don't want to delete the source list!
      call deleteBlock(set, blockNumber)
    end do

  end subroutine deleteRegionExtensions

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  !---------------------------------------------------------------------
  subroutine test
  !---------------------------------------------------------------------

    implicit none 

    real(double) :: myOffAxisAngles(2)
    type(SpacecraftAttitudeType) :: attitude, attBkp
    ! type defined in caloalutils/src/caloaldefs.f90
    type(SkyPositionType) :: skyPos, newSkyPos, oldSkyPos

    real(double) :: oldOffAxisAngles(2)

    !-------------------------------------------------------------------
    ! Start:

    call CAL_setState(instrument=RGS2)

attitude%ra   =  2.1369779898568871
attitude%dec  = -0.8262013799533605
attitude%apos =  5.0222221375344533

attBkp = attitude

skyPos%ra  =  2.1369928906537610
skyPos%dec = -0.8262388705572767
!skyPos%ra  =  attitude%ra
!skyPos%dec = attitude%dec

    myOffAxisAngles = raDecToOffAxis(attitude, skyPos) 
print *, ' my offAxisAngles:', myOffAxisAngles


oldOffAxisAngles = offAxisAngles(SpacecraftAttitude=attitude, SkyPos=skyPos)
print *, 'old offAxisAngles:', oldOffAxisAngles


attBkp = attitude
    newSkyPos = offAxisToRaDec(attBkp, myOffAxisAngles)

print *,    skyPos%ra,    skyPos%dec
print *, newSkyPos%ra, newSkyPos%dec
print *, ' '

attBkp = attitude
    oldSkyPos = offAxisToRaDec(attBkp, oldOffAxisAngles)
print *,    skyPos%ra,    skyPos%dec
print *, oldSkyPos%ra, oldSkyPos%dec

  end subroutine test

  !---------------------------------------------------------------------
  subroutine dumpRegion(x, y, fileName)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in) :: x(:), y(:)
    character(len=*), intent(in) ::&
      fileName

    integer :: n, i, ioDevNum
    character(len=255) ::&
      polygonStr

    !-------------------------------------------------------------------
    ! Start:

    n = size(x)

    write(polygonStr, '(a)') 'fk5;polygon('
    do i = 1, n - 1
      write(polygonStr, '(a, es13.6, a, es13.6, a)') trim(polygonStr)&
      , x(i), ',', y(i), ','
    end do
    write(polygonStr, '(a, es13.6, a, es13.6, a)') trim(polygonStr)&
    , x(n), ',', y(n), ')'

!***    call get_free_io_unit(ioDevNum)
ioDevNum = 1
    open(unit = ioDevNum, file = trim(fileName), status = 'replace')
      write(ioDevNum, '(a)') '# Region file format: DS9 version 3.0'
      write(ioDevNum, '(a)') "global color=green font='helvetica 10 &
      &normal' edit=1 move=1 delete=1 include=1 fixed=0"
      write(ioDevNum, '(a)') trim(polygonStr)
    close(ioDevNum)

  end subroutine dumpRegion

  !---------------------------------------------------------------------
  function calcAlpha(beta, y) result(alpha)
  !---------------------------------------------------------------------
  ! Uses the dispersion relation to calculate alpha, the angle between
  ! the incident ray and the grating, from beta, the angle between the
  ! dispersed ray and the grating, and y = order * wavelength
  ! / ruling_separation. Beta and alpha are assumed small and therefore
  ! a relation is used that avoids imprecision due to taking cosines of
  ! small angles.

    implicit none 

    real(single), intent(in) :: beta
    real(double), intent(in) :: y
    real(single) :: alpha

    !-------------------------------------------------------------------
    ! Start:

!    alpha = acos(cos(beta) - y) ! imprecise for small beta/alpha.
    alpha = asin(sqrt(sin(beta) * sin(beta) - y * y + 2.0 * y * cos(beta)))

  end function calcAlpha

  !---------------------------------------------------------------------
  subroutine dumpSrcData(srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(sourceType), dimension(numSrcs), intent(in) ::&
      srcData
    integer :: src

    !-------------------------------------------------------------------
    ! Start:

    do src = 1, numSrcs
      print *, ''
      print *, ''
      print *, 'Source ', src, ':'
      print *, ''
      print *, '       index:', srcData(src)%index
      print *, '       label:', srcData(src)%label
      print *, '          ra:', srcData(src)%raDeg
      print *, '         dec:', srcData(src)%decDeg
      print *, '        rate:', srcData(src)%rate
      print *, '  delta_disp:', srcData(src)%deltaDispArcmin
      print *, '  delta_xdsp:', srcData(src)%deltaXdspArcmin
      print *, '     fov_phi:', srcData(src)%fovPhiDeg
      print *, '       fov_r:', srcData(src)%fovRdeg
      print *, '   confusion:', srcData(src)%confusion
      print *, '     process:', srcData(src)%process
      print *, ' bkg_exclude:', srcData(src)%bkgExclude
      print *, 'fixed_on_sky:', srcData(src)%fixedOnSky
      print *, '    epic_num:', srcData(src)%epicFileAttrNum
    end do

  end subroutine dumpSrcData

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module rgssources_mod
