!+
! Module: rgssources_shared
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines shared between rgssources and rgssources_update.
! 
! $Id: rgssources_shared.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_shared

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use errorHandling
  use oal

  use pointing_mod

  use constants
!use temp_ssclib ! for intToStr
  use ssc_misc		! for intToStr

  use rgssources_init

  implicit none 

  integer(int16) :: formatNum
  character(20), dimension(0:1) :: tableNames = (/'SOURCES', 'SRCLIST'/)

  private

!  public getInstrument
!  public getExpId
!  public getPointing
!  public calcMaxNumSrcs
!  public fillOutputSrcList
!  public writeToDataSet
!  public flagCheck
!  public calcPrimeIndex

  public calculateFOVcoords
  public invertFOVcoords
  public calcRgsPlaneConfusion
  public writeNewTable
  public writeGenericAttributes
  public writeObservationAttributes
  public writeExposureAttributes

  public formatNum, tableNames

contains 

  !---------------------------------------------------------------------
  subroutine calculateFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    type(sourceType), intent(inout) ::&
      srcData
    type(SkyPositionType) :: srcSkyPos
    real(double) :: offAxisAngles(2)
integer(int16) :: status

    !-------------------------------------------------------------------
    ! Start:

    srcSkyPos%ra  = srcData%raDeg  * piOn180
    srcSkyPos%dec = srcData%decDeg * piOn180

!    offAxisAngles = raDecToOffAxis(attitude, srcSkyPos)
call raDecToOffAxis_new(attitude, srcSkyPos, offAxisAngles, status) ! (new in
    ! routine from rgslib/pointing_mod. This in turn calls
    ! rgslib/dummycal_mod function RGSLIB_getFoverL, which needs a CCF.

if (status .lt. 0) then
srcData%deltaDispArcmin = myReal32null
srcData%deltaXdspArcmin = myReal32null
srcData%fovPhiDeg       = myReal32null
srcData%fovRdeg         = myReal32null

srcData%flag = ior(srcData%flag, WIDESRC_MASK)

else
    srcData%deltaDispArcmin = offAxisAngles(1) * 60.0 / piOn180
    srcData%deltaXdspArcmin = offAxisAngles(2) * 60.0 / piOn180

    call dispXdspToPolar(&
      srcData%deltaDispArcmin,&
      srcData%deltaXdspArcmin,&
      srcData%fovPhiDeg,&
      srcData%fovRdeg&
    )
end if

  end subroutine calculateFOVcoords

  !---------------------------------------------------------------------
  subroutine invertFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    type(sourceType), intent(inout) ::&
      srcData
    type(SkyPositionType) :: srcSkyPos
    real(double) :: offAxisAngles(2)

    !-------------------------------------------------------------------
    ! Start:

    offAxisAngles(1) = srcData%deltaDispArcmin * piOn180 / 60.0
    offAxisAngles(2) = srcData%deltaXdspArcmin * piOn180 / 60.0

    srcSkyPos = offAxisToRaDec(attitude, offAxisAngles) ! new in
    ! rgslib-1.14
    ! routine from rgslib/pointing_mod. This in turn calls
    ! rgslib/dummycal_mod function RGSLIB_getFoverL, which needs a CCF.

    srcData%raDeg  = srcSkyPos%ra  / piOn180
    srcData%decDeg = srcSkyPos%dec / piOn180

  end subroutine invertFOVcoords

  !---------------------------------------------------------------------
  subroutine calcRgsPlaneConfusion(srcData, primeSrc)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      srcData
    integer(int16), intent(in) ::&
      primeSrc

    integer(int16) ::&
      numSrcs
    integer(int16) ::&
      src
    real(single) :: delta_disp, delta_xdisp, iPrime
 
!****** these should come from CAL (CCF MiscData)
    real, parameter :: p1 = -0.00653
    real, parameter :: p2 =  1.01
    real, parameter :: a1 = 36.9
    real, parameter :: a2 = 593.4
    real, parameter :: b1 = 2.29
    real, parameter :: b2 = 23.3

    ! See package documentation for equation 
    ! offaxis angle converted to arcsec

character(10) :: primeStr
!integer :: status

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating RGS confusion value.')

    numSrcs = size(srcData)

    if (srcData(primeSrc)%rate .le. 0.0) then
!call intToStr(int(primeSrc), 'left', primeStr, status)
primeStr = intToStr(int(primeSrc), leftJustify=.true._bool)
write(errstr, '(a)') 'The prime source (INDEX='//trim(primeStr)&
//') has a RATE value of <= zero. Confusion values cannot be calculated.'
call warning('badPrimeRate')

      srcData%confusion = 0.0

    else
      do src = 1, numSrcs
        delta_disp  = abs(srcData(src)%deltaDispArcmin&
        - srcData(primeSrc)%deltaDispArcmin)
        delta_xdisp = abs(srcData(src)%deltaXdspArcmin&
        - srcData(primeSrc)%deltaXdspArcmin)

        iPrime = p1 + p2 * srcData(src)%rate&
        * exp(-0.5*((delta_disp  * 60.0 + a1) / a2)**2)&
        * exp(-0.5*((delta_xdisp * 60.0 + b1) / b2)**2)
        srcData(src)%confusion = iPrime / srcData(primeSrc)%rate
      end do
    end if

  end subroutine calcRgsPlaneConfusion

  !---------------------------------------------------------------------
  subroutine writeNewTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(in) ::&
      srcData

    integer(int16) ::&
      numSrcs, src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol,&
      flagCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr,&
      flagColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr
    integer :: labelDims(1)

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    labelDims(1) = int(labelsize)

    ! Set up columns:
    indexCol = addColumn(tab, 'INDEX',        INTEGER16)
    labelCol = addColumn(tab, 'LABEL',        STRING, dimensions=labelDims&
      , comment = 'This string is unique for each source.')
    raCol    = addColumn(tab, 'RA',           REAL32, 'degrees')
    decCol   = addColumn(tab, 'DEC',          REAL32, 'degrees')
    rateCol  = addColumn(tab, 'RATE',         REAL32, 'cts/s'&
      , comment = 'Taken from EPIC source list, or user-supplied.')
    dispCol  = addColumn(tab, 'DELTA_DISP',   REAL32, 'arcmin'&
      , comment = 'Offset from pointing: dispersion direction.')
    xdspCol  = addColumn(tab, 'DELTA_XDSP',   REAL32, 'arcmin'&
      , comment = 'Offset from pointing: cross-dispersion direction.')
    phiCol   = addColumn(tab, 'FOV_PHI',      REAL32, 'degrees'&
      , comment = 'Offset from pointing in polar coordinates: angle from &
      &the -ve dispersion axis towards the +ve cross-dispersion axis.')
    rCol     = addColumn(tab, 'FOV_R',        REAL32, 'degrees'&
      , comment = 'Offset from pointing in polar coordinates: radius from &
      &pointing.')
    confnCol = addColumn(tab, 'CONFUSION',    REAL32&
      , comment = 'Fraction giving a measure of the confusion between each &
      &source and the prime source.')
    procCol  = addColumn(tab, 'PROCESS',      BOOLEAN&
      , comment = 'True means a spectrum extraction region should be &
      &constructed for this source.')
    excluCol = addColumn(tab, 'BKG_EXCLUDE',  BOOLEAN&
      , comment = 'True means this source should be excluded from the &
      &background spectrum extraction region.')
    fixedCol = addColumn(tab, 'FIXED_ON_SKY', BOOLEAN&
      , comment = 'True for sources whose positions were initially entered &
      &as ra and dec.')
    epicnCol = addColumn(tab, 'EPIC_FILE',    INTEGER32&
      , comment = 'Key to identify the appropriate EPICFILEn and &
      &EPICBANDn attributes.')
    flagCol  = addColumn(tab, 'FLAG',         INTEGER32&
      , comment = 'Quality flag.')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)
    flagColPtr  => int32Data(  flagCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', 'Could not set an int32 null value for &
&column EPIC_FILE.')
    end if

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate

      if (srcData(src)%deltaDispArcmin .eq. myReal32null) then
        call setToNull(dispCol, src - 1)
      else
        dispColPtr(src - 1) = srcData(src)%deltaDispArcmin
      end if

      if (srcData(src)%deltaXdspArcmin .eq. myReal32null) then
        call setToNull(xdspCol, src - 1)
      else
        xdspColPtr(src - 1) = srcData(src)%deltaXdspArcmin
      end if

      if (srcData(src)%fovPhiDeg .eq. myReal32null) then
        call setToNull(phiCol, src - 1)
      else
        phiColPtr(src - 1) = srcData(src)%fovPhiDeg
      end if

      if (srcData(src)%fovRdeg .eq. myReal32null) then
        call setToNull(rCol, src - 1)
      else
        rColPtr(src - 1) = srcData(src)%fovRdeg
      end if

      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky
      flagColPtr(src - 1)  = srcData(src)%flag

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

      call setStringCell(labelCol , src - 1, srcData(src)%label)
! DON'T trim the string argument to setStringCell. If you do, the characters
! in the file that should be trailing blanks will not be overwritten from
! whatever they were before in the memory map of the file, and will thus
! contain nonsense characters.
    end do

    call release(indexCol)
    call release(labelCol)
    call release(raCol)
    call release(decCol)
    call release(rateCol)
    call release(dispCol)
    call release(xdspCol)
    call release(phiCol)
    call release(rCol)
    call release(confnCol)
    call release(procCol)
    call release(excluCol)
    call release(fixedCol)
    call release(epicnCol)
    call release(flagCol)

  end subroutine writeNewTable

  !---------------------------------------------------------------------
  subroutine writeGenericAttributes(set, instrumentName, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    character(*), intent(in) :: instrumentName
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(set, 'TELESCOP', 'XMM')
    call setAttribute(set, 'INSTRUME', trim(instrumentName))
    call setAttribute(set, 'FILTER', 'UNKNOWN'&
    , comment='RGS is filterless.')
    call setAttribute(set, 'ORIGIN', 'Leicester/SSC'&
    , comment="Origin of FITS file")
    call setAttribute(set, 'RA_PNT',  attitude%ra   / piOn180)
    call setAttribute(set, 'DEC_PNT', attitude%dec  / piOn180)
    call setAttribute(set, 'PA_PNT',  attitude%apos / piOn180)
    call setAttribute(set, 'RADECSYS', 'FK5')
    call setAttribute(set, 'EQUINOX', 2000.0)
!    call setAttribute(set, 'EXPOSURE',)
!    call setAttribute(set, 'DATE-OBS',)
!    call setAttribute(set, 'DATE-END',)
!    call setAttribute(set, 'OCB',) ! special RGS header kwd: on-chip binning

  end subroutine writeGenericAttributes

  !---------------------------------------------------------------------
  subroutine writeObservationAttributes(set)
  !---------------------------------------------------------------------

    use oal

    implicit none 

    type(DataSetT), intent(in) :: set

    type(ProposalInfoType) :: proposalInfo
    type(OdfInfoType)      :: odfInfo

    !-------------------------------------------------------------------
    ! Start:

    call OAL_odfInfo(odfInfo) 
    call OAL_proposalInfo(proposalInfo)

    if (odfInfo%isSlew) then
      call setAttribute(set, "OBS_MODE", "SLEW"&
      , comment="Observation mode (pointing or slew)")
    else
      call setAttribute(set, "OBS_MODE", "POINTING"&
      , comment="Observation mode (pointing or slew)")
    endif
    call setAttribute(set, 'OBS_ID', trim(odfInfo%proposalId)&
    //trim(odfInfo%observationId), comment="Observation identifier")
    call setAttribute(set, "OBJECT", trim(proposalInfo%targetName)&
    , comment="Name of observed object")
    call setAttribute(set, "OBSERVER", trim(proposalInfo%GoName)&
    , comment="Name of observer")

  end subroutine writeObservationAttributes

  !---------------------------------------------------------------------
!  subroutine writeExposureAttributes(set, instrumentId, exposureIdStr)
subroutine writeExposureAttributes(set, instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    use oal
use caloaldefs
use odfdesignators

    implicit none 

    type(DataSetT), intent(in) :: set
    integer(int32), intent(in) :: instrumentId
!    character(4),   intent(in) :: exposureIdStr
character(*),   intent(in) :: instExpIdStr

    type(OdfInfoType)      :: odfInfo
    type(ExposureInfoType) :: exposureInfo
    integer(int32)         :: dataMode
    character(13)          :: obsExpIdStr
    character(10)          :: obsIdStr
character(4)           :: fullExpIdStr
type(InstrumentExposureIdentifierT) :: instExpIdStructure
! type defined in caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start:

    call OAL_odfInfo(odfInfo) 
    obsIdStr = odfInfo%proposalId(1:6)//odfInfo%observationId(1:4)
!    call OAL_setState(instrument=instrumentId, exposureId=exposureIdStr)
call OAL_setState(instrument=instrumentId, instrExposureId=instExpIdStr)
    call OAL_getState(dataMode=dataMode)
    call OAL_exposureInfo(exposureInfo) 

    call setAttribute(set, 'DATAMODE', dataModeId(dataMode)&
    , comment='Instrument mode.')
    write(obsExpIdStr, '(a10, i3.3)') obsIdStr, exposureInfo%number
    call setAttribute(set, 'EXP_ID', obsExpIdStr&
    , comment="Exposure identifier")
!    call setAttribute(set, 'EXPIDSTR', exposureIdStr&
instExpIdStructure = instrumentExposureIdentifier(trim(instExpIdStr))
fullExpIdStr = exposureId(instExpIdStructure)
call setAttribute(set, 'EXPIDSTR', fullExpIdStr&
    , comment="Scheduled/unscheduled letter plus exposure number.")

  end subroutine writeExposureAttributes

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template
end module rgssources_shared






